@@ -23,51 +23,132 @@ import hashlib
 import logging
 import os
 import stat
-import urlparse
 
 import jsonschema
-from oslo.serialization import jsonutils
 from oslo_config import cfg
+from oslo_serialization import jsonutils
+from oslo_utils import encodeutils
 from oslo_utils import excutils
 from oslo_utils import units
+from six.moves import urllib
 
 import glance_store
 from glance_store import capabilities
 from glance_store.common import utils
 import glance_store.driver
 from glance_store import exceptions
-from glance_store import i18n
+from glance_store.i18n import _, _LE, _LW
 import glance_store.location
 
 
 LOG = logging.getLogger(__name__)
-_ = i18n._
-_LE = i18n._LE
-_LW = i18n._LW
 
 _FILESYSTEM_CONFIGS = [
     cfg.StrOpt('filesystem_store_datadir',
-               help=_('Directory to which the Filesystem backend '
-                      'store writes images.')),
+               default='/var/lib/glance/images',
+               help=_("""
+Directory to which the filesystem backend store writes images.
+
+Upon start up, Glance creates the directory if it doesn't already
+exist and verifies write access to the user under which
+``glance-api`` runs. If the write access isn't available, a
+``BadStoreConfiguration`` exception is raised and the filesystem
+store may not be available for adding new images.
+
+NOTE: This directory is used only when filesystem store is used as a
+storage backend. Either ``filesystem_store_datadir`` or
+``filesystem_store_datadirs`` option must be specified in
+``glance-api.conf``. If both options are specified, a
+``BadStoreConfiguration`` will be raised and the filesystem store
+may not be available for adding new images.
+
+Possible values:
+    * A valid path to a directory
+
+Related options:
+    * ``filesystem_store_datadirs``
+    * ``filesystem_store_file_perm``
+
+""")),
     cfg.MultiStrOpt('filesystem_store_datadirs',
-                    help=_("List of directories and its priorities to which "
-                           "the Filesystem backend store writes images.")),
+                    help=_("""
+List of directories and their priorities to which the filesystem
+backend store writes images.
+
+The filesystem store can be configured to store images in multiple
+directories as opposed to using a single directory specified by the
+``filesystem_store_datadir`` configuration option. When using
+multiple directories, each directory can be given an optional
+priority to specify the preference order in which they should
+be used. Priority is an integer that is concatenated to the
+directory path with a colon where a higher value indicates higher
+priority. When two directories have the same priority, the directory
+with most free space is used. When no priority is specified, it
+defaults to zero.
+
+More information on configuring filesystem store with multiple store
+directories can be found at
+http://docs.openstack.org/developer/glance/configuring.html
+
+NOTE: This directory is used only when filesystem store is used as a
+storage backend. Either ``filesystem_store_datadir`` or
+``filesystem_store_datadirs`` option must be specified in
+``glance-api.conf``. If both options are specified, a
+``BadStoreConfiguration`` will be raised and the filesystem store
+may not be available for adding new images.
+
+Possible values:
+    * List of strings of the following form:
+        * ``<a valid directory path>:<optional integer priority>``
+
+Related options:
+    * ``filesystem_store_datadir``
+    * ``filesystem_store_file_perm``
+
+""")),
     cfg.StrOpt('filesystem_store_metadata_file',
-               help=_("The path to a file which contains the "
-                      "metadata to be returned with any location "
-                      "associated with this store.  The file must "
-                      "contain a valid JSON object. The object should contain "
-                      "the keys 'id' and 'mountpoint'. The value for both "
-                      "keys should be 'string'.")),
+               help=_("""
+Filesystem store metadata file.
+
+The path to a file which contains the metadata to be returned with
+any location associated with the filesystem store. The file must
+contain a valid JSON object. The object should contain the keys
+``id`` and ``mountpoint``. The value for both keys should be a
+string.
+
+Possible values:
+    * A valid path to the store metadata file
+
+Related options:
+    * None
+
+""")),
     cfg.IntOpt('filesystem_store_file_perm',
                default=0,
-               help=_("The required permission for created image file. "
-                      "In this way the user other service used, e.g. Nova, "
-                      "who consumes the image could be the exclusive member "
-                      "of the group that owns the files created. Assigning "
-                      "it less then or equal to zero means don't change the "
-                      "default permission of the file. This value will be "
-                      "decoded as an octal digit."))]
+               help=_("""
+File access permissions for the image files.
+
+Set the intended file access permissions for image data. This provides
+a way to enable other services, e.g. Nova, to consume images directly
+from the filesystem store. The users running the services that are
+intended to be given access to could be made a member of the group
+that owns the files created. Assigning a value less then or equal to
+zero for this configuration option signifies that no changes be made
+to the  default permissions. This value will be decoded as an octal
+digit.
+
+For more information, please refer the documentation at
+http://docs.openstack.org/developer/glance/configuring.html
+
+Possible values:
+    * A valid file access permission
+    * Zero
+    * Any negative integer
+
+Related options:
+    * None
+
+"""))]
 
 MULTI_FILESYSTEM_METADATA_SCHEMA = {
     "type": "array",
@@ -98,7 +179,7 @@ class StoreLocation(glance_store.location.StoreLocation):
         in the URL are interpreted differently in Python 2.6.1+ than prior
         versions of Python.
         """
-        pieces = urlparse.urlparse(uri)
+        pieces = urllib.parse.urlparse(uri)
         assert pieces.scheme in ('file', 'filesystem')
         self.scheme = pieces.scheme
         path = (pieces.netloc + pieces.path).strip()
@@ -175,7 +256,7 @@ class Store(glance_store.driver.Store):
         write permission.
 
         :datadir is a directory path in which glance wites image files.
-        :raise BadStoreConfiguration exception if datadir is read-only.
+        :raises: BadStoreConfiguration exception if datadir is read-only.
         """
         if not os.access(datadir, os.W_OK):
             msg = (_("Permission to write in %s denied") % datadir)
@@ -209,8 +290,9 @@ class Store(glance_store.driver.Store):
                     # chmod g+x
                     os.chmod(datadir, mode | stat.S_IXGRP)
         except (IOError, OSError):
-            LOG.warn(_LW("Unable to set execution permission of owner-group "
-                         "and/or other-users to datadir: %s") % datadir)
+            LOG.warning(_LW("Unable to set execution permission of "
+                            "owner-group and/or other-users to datadir: %s")
+                        % datadir)
 
     def _create_image_directories(self, directory_paths):
         """
@@ -218,7 +300,7 @@ class Store(glance_store.driver.Store):
         it does not exist.
 
         :directory_paths is a list of directories belonging to glance store.
-        :raise BadStoreConfiguration exception if creating a directory fails.
+        :raises: BadStoreConfiguration exception if creating a directory fails.
         """
         for datadir in directory_paths:
             if os.path.exists(datadir):
@@ -256,7 +338,7 @@ class Store(glance_store.driver.Store):
         :raises: BadStoreConfiguration exception if metadata is not valid.
         """
         try:
-            with open(metadata_file, 'r') as fptr:
+            with open(metadata_file, 'rb') as fptr:
                 metadata = jsonutils.load(fptr)
 
             if isinstance(metadata, dict):
@@ -271,18 +353,20 @@ class Store(glance_store.driver.Store):
             self.FILESYSTEM_STORE_METADATA = metadata
         except (jsonschema.exceptions.ValidationError,
                 exceptions.BackendException, ValueError) as vee:
+            err_msg = encodeutils.exception_to_unicode(vee)
             reason = _('The JSON in the metadata file %(file)s is '
                        'not valid and it can not be used: '
                        '%(vee)s.') % dict(file=metadata_file,
-                                          vee=utils.exception_to_str(vee))
+                                          vee=err_msg)
             LOG.error(reason)
             raise exceptions.BadStoreConfiguration(
                 store_name="filesystem", reason=reason)
         except IOError as ioe:
+            err_msg = encodeutils.exception_to_unicode(ioe)
             reason = _('The path for the metadata file %(file)s could '
                        'not be accessed: '
                        '%(ioe)s.') % dict(file=metadata_file,
-                                          ioe=utils.exception_to_str(ioe))
+                                          ioe=err_msg)
             LOG.error(reason)
             raise exceptions.BadStoreConfiguration(
                 store_name="filesystem", reason=reason)
@@ -294,8 +378,8 @@ class Store(glance_store.driver.Store):
         this method. If the store was not able to successfully configure
         itself, it should raise `exceptions.BadStoreConfiguration`
         """
-        if not (self.conf.glance_store.filesystem_store_datadir
-                or self.conf.glance_store.filesystem_store_datadirs):
+        if not (self.conf.glance_store.filesystem_store_datadir or
+                self.conf.glance_store.filesystem_store_datadirs):
             reason = (_("Specify at least 'filesystem_store_datadir' or "
                         "'filesystem_store_datadirs' option"))
             LOG.error(reason)
@@ -335,10 +419,12 @@ class Store(glance_store.driver.Store):
             for datadir in self.conf.glance_store.filesystem_store_datadirs:
                 (datadir_path,
                  priority) = self._get_datadir_path_and_priority(datadir)
-                self._check_directory_paths(datadir_path, directory_paths)
+                priority_paths = self.priority_data_map.setdefault(
+                    int(priority), [])
+                self._check_directory_paths(datadir_path, directory_paths,
+                                            priority_paths)
                 directory_paths.add(datadir_path)
-                self.priority_data_map.setdefault(int(priority),
-                                                  []).append(datadir_path)
+                priority_paths.append(datadir_path)
 
             self.priority_list = sorted(self.priority_data_map,
                                         reverse=True)
@@ -349,13 +435,14 @@ class Store(glance_store.driver.Store):
         if metadata_file:
             self._validate_metadata(metadata_file)
 
-    def _check_directory_paths(self, datadir_path, directory_paths):
+    def _check_directory_paths(self, datadir_path, directory_paths,
+                               priority_paths):
         """
         Checks if directory_path is already present in directory_paths.
 
         :datadir_path is directory path.
         :datadir_paths is set of all directory paths.
-        :raise BadStoreConfiguration exception if same directory path is
+        :raises: BadStoreConfiguration exception if same directory path is
                already present in directory_paths.
         """
         if datadir_path in directory_paths:
@@ -363,23 +450,29 @@ class Store(glance_store.driver.Store):
                      "multiple times in filesystem_store_datadirs "
                      "option of filesystem configuration") %
                    {'datadir_path': datadir_path})
-            LOG.exception(msg)
-            raise exceptions.BadStoreConfiguration(
-                store_name="filesystem", reason=msg)
+
+            # If present with different priority it's a bad configuration
+            if datadir_path not in priority_paths:
+                LOG.exception(msg)
+                raise exceptions.BadStoreConfiguration(
+                    store_name="filesystem", reason=msg)
+
+            # Present with same prio (exact duplicate) only deserves a warning
+            LOG.warning(msg)
 
     def _get_datadir_path_and_priority(self, datadir):
         """
         Gets directory paths and its priority from
         filesystem_store_datadirs option in glance-api.conf.
 
-        :datadir is directory path with its priority.
-        :returns datadir_path as directory path
+        :param datadir: is directory path with its priority.
+        :returns: datadir_path as directory path
                  priority as priority associated with datadir_path
-        :raise BadStoreConfiguration exception if priority is invalid or
+        :raises: BadStoreConfiguration exception if priority is invalid or
                empty directory path is specified.
         """
         priority = 0
-        parts = map(lambda x: x.strip(), datadir.rsplit(":", 1))
+        parts = [part.strip() for part in datadir.rsplit(":", 1)]
         datadir_path = parts[0]
         if len(parts) == 2 and parts[1]:
             priority = parts[1]
@@ -451,9 +544,9 @@ class Store(glance_store.driver.Store):
         where to find the image file, and returns a tuple of generator
         (for reading the image file) and image_size
 
-        :param location `glance_store.location.Location` object, supplied
+        :param location: `glance_store.location.Location` object, supplied
                         from glance_store.location.get_location_from_uri()
-        :raises `glance_store.exceptions.NotFound` if image does not exist
+        :raises: `glance_store.exceptions.NotFound` if image does not exist
         """
         filepath, filesize = self._resolve_location(location)
         msg = _("Found image at %s. Returning in ChunkedFile.") % filepath
@@ -469,10 +562,10 @@ class Store(glance_store.driver.Store):
         Takes a `glance_store.location.Location` object that indicates
         where to find the image file and returns the image size
 
-        :param location `glance_store.location.Location` object, supplied
+        :param location: `glance_store.location.Location` object, supplied
                         from glance_store.location.get_location_from_uri()
-        :raises `glance_store.exceptions.NotFound` if image does not exist
-        :rtype int
+        :raises: `glance_store.exceptions.NotFound` if image does not exist
+        :rtype: int
         """
         filepath, filesize = self._resolve_location(location)
         msg = _("Found image at %s.") % filepath
@@ -485,11 +578,11 @@ class Store(glance_store.driver.Store):
         Takes a `glance_store.location.Location` object that indicates
         where to find the image file to delete
 
-        :location `glance_store.location.Location` object, supplied
+        :param location: `glance_store.location.Location` object, supplied
                   from glance_store.location.get_location_from_uri()
 
-        :raises NotFound if image does not exist
-        :raises Forbidden if cannot delete because of permissions
+        :raises: NotFound if image does not exist
+        :raises: Forbidden if cannot delete because of permissions
         """
         loc = location.store_location
         fn = loc.path
@@ -498,7 +591,8 @@ class Store(glance_store.driver.Store):
                 LOG.debug(_("Deleting image at %(fn)s"), {'fn': fn})
                 os.unlink(fn)
             except OSError:
-                raise exceptions.Forbidden(_("You cannot delete file %s") % fn)
+                raise exceptions.Forbidden(
+                    message=(_("You cannot delete file %s") % fn))
         else:
             raise exceptions.NotFound(image=fn)
 
@@ -520,9 +614,9 @@ class Store(glance_store.driver.Store):
         free space, in priority order. If two suitable directories have
         the same priority, choose the one with the most free space
         available.
-        :image_size size of image being uploaded.
-        :returns best_datadir as directory path of the best priority datadir.
-        :raises exceptions.StorageFull if there is no datadir in
+        :param image_size: size of image being uploaded.
+        :returns: best_datadir as directory path of the best priority datadir.
+        :raises: exceptions.StorageFull if there is no datadir in
                 self.priority_data_map that can accommodate the image.
         """
         if not self.multiple_datadirs:
@@ -551,7 +645,8 @@ class Store(glance_store.driver.Store):
         return best_datadir
 
     @capabilities.check
-    def add(self, image_id, image_file, image_size, context=None):
+    def add(self, image_id, image_file, image_size, context=None,
+            verifier=None):
         """
         Stores an image file with supplied identifier to the backend
         storage system and returns a tuple containing information
@@ -560,13 +655,14 @@ class Store(glance_store.driver.Store):
         :param image_id: The opaque image identifier
         :param image_file: The image data to write, as a file-like object
         :param image_size: The size of the image data to write, in bytes
+        :param verifier: An object used to verify signatures for images
 
-        :retval tuple of URL in backing store, bytes written, checksum
+        :retval: tuple of URL in backing store, bytes written, checksum
                 and a dictionary with storage system specific information
-        :raises `glance_store.exceptions.Duplicate` if the image already
+        :raises: `glance_store.exceptions.Duplicate` if the image already
                 existed
 
-        :note By default, the backend writes the image data to a file
+        :note:: By default, the backend writes the image data to a file
               `/<DATADIR>/<ID>`, where <DATADIR> is the value of
               the filesystem_store_datadir configuration option and <ID>
               is the supplied image ID.
@@ -586,6 +682,8 @@ class Store(glance_store.driver.Store):
                                                self.WRITE_CHUNKSIZE):
                     bytes_written += len(buf)
                     checksum.update(buf)
+                    if verifier:
+                        verifier.update(buf)
                     f.write(buf)
         except IOError as e:
             if e.errno != errno.EACCES:
@@ -613,8 +711,8 @@ class Store(glance_store.driver.Store):
             try:
                 os.chmod(filepath, perm)
             except (IOError, OSError):
-                LOG.warn(_LW("Unable to set permission to image: %s") %
-                         filepath)
+                LOG.warning(_LW("Unable to set permission to image: %s") %
+                            filepath)
 
         return ('file://%s' % filepath, bytes_written, checksum_hex, metadata)
 
@@ -625,4 +723,5 @@ class Store(glance_store.driver.Store):
         except Exception as e:
             msg = _('Unable to remove partial image '
                     'data for image %(iid)s: %(e)s')
-            LOG.error(msg % dict(iid=iid, e=utils.exception_to_str(e)))
+            LOG.error(msg % dict(iid=iid,
+                                 e=encodeutils.exception_to_unicode(e)))
