@@ -13,8 +13,7 @@
 #    under the License.
 
 import logging
-
-import collections
+import sys
 
 from oslo_config import cfg
 from six.moves import configparser
@@ -105,10 +104,10 @@ _config_defaults = {'user_domain_id': 'default',
                     'project_domain_id': 'default',
                     'project_domain_name': None}
 
-# NOTE(bourke): The default dict_type is collections.OrderedDict in py27, but
-# we must set manually for compatibility with py26
-CONFIG = configparser.SafeConfigParser(defaults=_config_defaults,
-                                       dict_type=collections.OrderedDict)
+if sys.version_info >= (3, 2):
+    CONFIG = configparser.ConfigParser(defaults=_config_defaults)
+else:
+    CONFIG = configparser.SafeConfigParser(defaults=_config_defaults)
 LOG = logging.getLogger(__name__)
 
 
