@@ -16,118 +16,235 @@
 """Storage backend for VMware Datastore"""
 
 import hashlib
-import httplib
 import logging
 import os
-import socket
 
 from oslo_config import cfg
 from oslo_utils import excutils
+from oslo_utils import netutils
 from oslo_utils import units
-from oslo_vmware import api
-from oslo_vmware import constants
-import oslo_vmware.exceptions as vexc
-from oslo_vmware.objects import datacenter as oslo_datacenter
-from oslo_vmware.objects import datastore as oslo_datastore
-from oslo_vmware import vim_util
+try:
+    from oslo_vmware import api
+    import oslo_vmware.exceptions as vexc
+    from oslo_vmware.objects import datacenter as oslo_datacenter
+    from oslo_vmware.objects import datastore as oslo_datastore
+    from oslo_vmware import vim_util
+except ImportError:
+    api = None
 
+from six.moves import urllib
+import six.moves.urllib.parse as urlparse
+
+import requests
+from requests import adapters
+from requests.packages.urllib3.util import retry
 import six
 # NOTE(jokke): simplified transition to py3, behaves like py2 xrange
 from six.moves import range
-import six.moves.urllib.parse as urlparse
 
 import glance_store
 from glance_store import capabilities
+from glance_store.common import utils
 from glance_store import exceptions
-from glance_store.i18n import _
-from glance_store.i18n import _LE
+from glance_store.i18n import _, _LE
 from glance_store import location
 
 
 LOG = logging.getLogger(__name__)
 
+CHUNKSIZE = 1024 * 64  # 64kB
 MAX_REDIRECTS = 5
 DEFAULT_STORE_IMAGE_DIR = '/openstack_glance'
 DS_URL_PREFIX = '/folder'
 STORE_SCHEME = 'vsphere'
 
-# check that datacenter/datastore combination is valid
-_datastore_info_valid = False
-
 _VMWARE_OPTS = [
-    cfg.StrOpt('vmware_server_host',
-               help=_('ESX/ESXi or vCenter Server target system. '
-                      'The server value can be an IP address or a DNS name.')),
+    cfg.HostAddressOpt('vmware_server_host',
+                       sample_default='127.0.0.1',
+                       help=_("""
+Address of the ESX/ESXi or vCenter Server target system.
+
+This configuration option sets the address of the ESX/ESXi or vCenter
+Server target system. This option is required when using the VMware
+storage backend. The address can contain an IP address (127.0.0.1) or
+a DNS name (www.my-domain.com).
+
+Possible Values:
+    * A valid IPv4 or IPv6 address
+    * A valid DNS name
+
+Related options:
+    * vmware_server_username
+    * vmware_server_password
+
+""")),
     cfg.StrOpt('vmware_server_username',
-               help=_('Username for authenticating with '
-                      'VMware ESX/VC server.')),
+               sample_default='root',
+               help=_("""
+Server username.
+
+This configuration option takes the username for authenticating with
+the VMware ESX/ESXi or vCenter Server. This option is required when
+using the VMware storage backend.
+
+Possible Values:
+    * Any string that is the username for a user with appropriate
+      privileges
+
+Related options:
+    * vmware_server_host
+    * vmware_server_password
+
+""")),
     cfg.StrOpt('vmware_server_password',
-               help=_('Password for authenticating with '
-                      'VMware ESX/VC server.'),
+               sample_default='vmware',
+               help=_("""
+Server password.
+
+This configuration option takes the password for authenticating with
+the VMware ESX/ESXi or vCenter Server. This option is required when
+using the VMware storage backend.
+
+Possible Values:
+    * Any string that is a password corresponding to the username
+      specified using the "vmware_server_username" option
+
+Related options:
+    * vmware_server_host
+    * vmware_server_username
+
+"""),
                secret=True),
-    cfg.StrOpt('vmware_datacenter_path',
-               default=constants.ESX_DATACENTER_PATH,
-               help=_('DEPRECATED. Inventory path to a datacenter. '
-                      'If the vmware_server_host specified is an ESX/ESXi, '
-                      'the vmware_datacenter_path is optional. If specified, '
-                      'it should be "ha-datacenter". This option is '
-                      'deprecated in favor of vmware_datastores and will be '
-                      'removed in the Liberty release.'),
-               deprecated_for_removal=True),
-    cfg.StrOpt('vmware_datastore_name',
-               help=_('DEPRECATED. Datastore associated with the datacenter. '
-                      'This option is deprecated in favor of '
-                      'vmware_datastores and will be removed in the Liberty '
-                      'release.'),
-               deprecated_for_removal=True),
     cfg.IntOpt('vmware_api_retry_count',
                default=10,
-               help=_('Number of times VMware ESX/VC server API must be '
-                      'retried upon connection related issues.')),
+               min=1,
+               help=_("""
+The number of VMware API retries.
+
+This configuration option specifies the number of times the VMware
+ESX/VC server API must be retried upon connection related issues or
+server API call overload. It is not possible to specify 'retry
+forever'.
+
+Possible Values:
+    * Any positive integer value
+
+Related options:
+    * None
+
+""")),
     cfg.IntOpt('vmware_task_poll_interval',
                default=5,
-               help=_('The interval used for polling remote tasks '
-                      'invoked on VMware ESX/VC server.')),
+               min=1,
+               help=_("""
+Interval in seconds used for polling remote tasks invoked on VMware
+ESX/VC server.
+
+This configuration option takes in the sleep time in seconds for polling an
+on-going async task as part of the VMWare ESX/VC server API call.
+
+Possible Values:
+    * Any positive integer value
+
+Related options:
+    * None
+
+""")),
     cfg.StrOpt('vmware_store_image_dir',
                default=DEFAULT_STORE_IMAGE_DIR,
-               help=_('The name of the directory where the glance images '
-                      'will be stored in the VMware datastore.')),
-    cfg.BoolOpt('vmware_api_insecure',
+               help=_("""
+The directory where the glance images will be stored in the datastore.
+
+This configuration option specifies the path to the directory where the
+glance images will be stored in the VMware datastore. If this option
+is not set,  the default directory where the glance images are stored
+is openstack_glance.
+
+Possible Values:
+    * Any string that is a valid path to a directory
+
+Related options:
+    * None
+
+""")),
+    cfg.BoolOpt('vmware_insecure',
                 default=False,
-                help=_('Allow to perform insecure SSL requests to ESX/VC.')),
+                deprecated_name='vmware_api_insecure',
+                help=_("""
+Set verification of the ESX/vCenter server certificate.
+
+This configuration option takes a boolean value to determine
+whether or not to verify the ESX/vCenter server certificate. If this
+option is set to True, the ESX/vCenter server certificate is not
+verified. If this option is set to False, then the default CA
+truststore is used for verification.
+
+This option is ignored if the "vmware_ca_file" option is set. In that
+case, the ESX/vCenter server certificate will then be verified using
+the file specified using the "vmware_ca_file" option .
+
+Possible Values:
+    * True
+    * False
+
+Related options:
+    * vmware_ca_file
+
+""")),
+    cfg.StrOpt('vmware_ca_file',
+               sample_default='/etc/ssl/certs/ca-certificates.crt',
+               help=_("""
+Absolute path to the CA bundle file.
+
+This configuration option enables the operator to use a custom
+Cerificate Authority File to verify the ESX/vCenter certificate.
+
+If this option is set, the "vmware_insecure" option will be ignored
+and the CA file specified will be used to authenticate the ESX/vCenter
+server certificate and establish a secure connection to the server.
+
+Possible Values:
+    * Any string that is a valid absolute path to a CA file
+
+Related options:
+    * vmware_insecure
+
+""")),
     cfg.MultiStrOpt(
         'vmware_datastores',
-        help=_(
-            'A list of datastores where the image can be stored. This option '
-            'may be specified multiple times for specifying multiple '
-            'datastores. Either one of vmware_datastore_name or '
-            'vmware_datastores is required. The datastore name should be '
-            'specified after its datacenter path, seperated by ":". An '
-            'optional weight may be given after the datastore name, seperated '
-            'again by ":". Thus, the required format becomes '
-            '<datacenter_path>:<datastore_name>:<optional_weight>. When '
-            'adding an image, the datastore with highest weight will be '
-            'selected, unless there is not enough free space available in '
-            'cases where the image size is already known. If no weight is '
-            'given, it is assumed to be zero and the directory will be '
-            'considered for selection last. If multiple datastores have the '
-            'same weight, then the one with the most free space available is '
-            'selected.'))]
+        help=_("""
+The datastores where the image can be stored.
 
+This configuration option specifies the datastores where the image can
+be stored in the VMWare store backend. This option may be specified
+multiple times for specifying multiple datastores. The datastore name
+should be specified after its datacenter path, separated by ":". An
+optional weight may be given after the datastore name, separated again
+by ":" to specify the priority. Thus, the required format becomes
+<datacenter_path>:<datastore_name>:<optional_weight>.
 
-def is_valid_ipv6(address):
-    try:
-        socket.inet_pton(socket.AF_INET6, address)
-        return True
-    except Exception:
-        return False
+When adding an image, the datastore with highest weight will be
+selected, unless there is not enough free space available in cases
+where the image size is already known. If no weight is given, it is
+assumed to be zero and the directory will be considered for selection
+last. If multiple datastores have the same weight, then the one with
+the most free space available is selected.
+
+Possible Values:
+    * Any string of the format:
+      <datacenter_path>:<datastore_name>:<optional_weight>
+
+Related options:
+   * None
+
+"""))]
 
 
 def http_response_iterator(conn, response, size):
     """Return an iterator for a file-like object.
 
     :param conn: HTTP(S) Connection
-    :param response: httplib.HTTPResponse object
+    :param response: http_client.HTTPResponse object
     :param size: Chunk size to iterate with
     """
     try:
@@ -141,15 +258,18 @@ def http_response_iterator(conn, response, size):
 
 class _Reader(object):
 
-    def __init__(self, data):
+    def __init__(self, data, verifier=None):
         self._size = 0
         self.data = data
         self.checksum = hashlib.md5()
+        self.verifier = verifier
 
     def read(self, size=None):
         result = self.data.read(size)
         self._size += len(result)
         self.checksum.update(result)
+        if self.verifier:
+            self.verifier.update(result)
         return result
 
     @property
@@ -157,42 +277,6 @@ class _Reader(object):
         return self._size
 
 
-class _ChunkReader(_Reader):
-
-    def __init__(self, data, blocksize=8192):
-        self.blocksize = blocksize
-        self.current_chunk = ""
-        self.closed = False
-        super(_ChunkReader, self).__init__(data)
-
-    def read(self, size=None):
-        ret = ""
-        while size is None or size >= len(self.current_chunk):
-            ret += self.current_chunk
-            if size is not None:
-                size -= len(self.current_chunk)
-            if self.closed:
-                self.current_chunk = ""
-                break
-            self._get_chunk()
-        else:
-            ret += self.current_chunk[:size]
-            self.current_chunk = self.current_chunk[size:]
-        return ret
-
-    def _get_chunk(self):
-        if not self.closed:
-            chunk = self.data.read(self.blocksize)
-            chunk_len = len(chunk)
-            self._size += chunk_len
-            self.checksum.update(chunk)
-            if chunk:
-                self.current_chunk = '%x\r\n%s\r\n' % (chunk_len, chunk)
-            else:
-                self.current_chunk = '0\r\n\r\n'
-                self.closed = True
-
-
 class StoreLocation(location.StoreLocation):
     """Class describing an VMware URI.
 
@@ -216,10 +300,10 @@ class StoreLocation(location.StoreLocation):
         param_list = {'dsName': self.datstore_name}
         if self.datacenter_path:
             param_list['dcPath'] = self.datacenter_path
-        self.query = urlparse.urlencode(param_list)
+        self.query = urllib.parse.urlencode(param_list)
 
     def get_uri(self):
-        if is_valid_ipv6(self.server_host):
+        if netutils.is_valid_ipv6(self.server_host):
             base_url = '%s://[%s]%s' % (self.scheme,
                                         self.server_host, self.path)
         else:
@@ -242,7 +326,7 @@ class StoreLocation(location.StoreLocation):
             LOG.info(reason)
             raise exceptions.BadStoreUri(message=reason)
         (self.scheme, self.server_host,
-         path, params, query, fragment) = urlparse.urlparse(uri)
+         path, params, query, fragment) = urllib.parse.urlparse(uri)
         if not query:
             path, query = path.split('?')
 
@@ -252,7 +336,7 @@ class StoreLocation(location.StoreLocation):
         # reason = 'Badly formed VMware datastore URI %(uri)s.' % {'uri': uri}
         # LOG.debug(reason)
         # raise exceptions.BadStoreUri(reason)
-        parts = urlparse.parse_qs(self.query)
+        parts = urllib.parse.parse_qs(self.query)
         dc_path = parts.get('dcPath')
         if dc_path:
             self.datacenter_path = dc_path[0]
@@ -260,6 +344,16 @@ class StoreLocation(location.StoreLocation):
         if ds_name:
             self.datastore_name = ds_name[0]
 
+    @property
+    def https_url(self):
+        """
+        Creates a https url that can be used to upload/download data from a
+        vmware store.
+        """
+        parsed_url = urlparse.urlparse(self.get_uri())
+        new_url = parsed_url._replace(scheme='https')
+        return urlparse.urlunparse(new_url)
+
 
 class Store(glance_store.Store):
     """An implementation of the VMware datastore adapter."""
@@ -268,21 +362,18 @@ class Store(glance_store.Store):
                      capabilities.BitMasks.DRIVER_REUSABLE)
     OPTIONS = _VMWARE_OPTS
     WRITE_CHUNKSIZE = units.Mi
-    # FIXME(arnaud): re-visit this code once the store API is cleaned up.
-    _VMW_SESSION = None
 
     def __init__(self, conf):
         super(Store, self).__init__(conf)
         self.datastores = {}
 
-    def reset_session(self, force=False):
-        if Store._VMW_SESSION is None or force:
-            Store._VMW_SESSION = api.VMwareAPISession(
-                self.server_host, self.server_username, self.server_password,
-                self.api_retry_count, self.tpoll_interval)
-        return Store._VMW_SESSION
-
-    session = property(reset_session)
+    def reset_session(self):
+        self.session = api.VMwareAPISession(
+            self.server_host, self.server_username, self.server_password,
+            self.api_retry_count, self.tpoll_interval,
+            cacert=self.ca_file,
+            insecure=self.api_insecure)
+        return self.session
 
     def get_schemes(self):
         return (STORE_SCHEME,)
@@ -300,23 +391,7 @@ class Store(glance_store.Store):
             raise exceptions.BadStoreConfiguration(
                 store_name='vmware_datastore', reason=msg)
 
-        if not (self.conf.glance_store.vmware_datastore_name
-                or self.conf.glance_store.vmware_datastores):
-            msg = (_("Specify at least 'vmware_datastore_name' or "
-                     "'vmware_datastores' option"))
-            LOG.error(msg)
-            raise exceptions.BadStoreConfiguration(
-                store_name='vmware_datastore', reason=msg)
-
-        if (self.conf.glance_store.vmware_datastore_name and
-                self.conf.glance_store.vmware_datastores):
-            msg = (_("Specify either 'vmware_datastore_name' or "
-                     "'vmware_datastores' option"))
-            LOG.error(msg)
-            raise exceptions.BadStoreConfiguration(
-                store_name='vmware_datastore', reason=msg)
-
-    def configure(self):
+    def configure(self, re_raise_bsc=False):
         self._sanity_check()
         self.scheme = STORE_SCHEME
         self.server_host = self._option_get('vmware_server_host')
@@ -324,8 +399,14 @@ class Store(glance_store.Store):
         self.server_password = self._option_get('vmware_server_password')
         self.api_retry_count = self.conf.glance_store.vmware_api_retry_count
         self.tpoll_interval = self.conf.glance_store.vmware_task_poll_interval
-        self.api_insecure = self.conf.glance_store.vmware_api_insecure
-        super(Store, self).configure()
+        self.ca_file = self.conf.glance_store.vmware_ca_file
+        self.api_insecure = self.conf.glance_store.vmware_insecure
+        if api is None:
+            msg = _("Missing dependencies: oslo_vmware")
+            raise exceptions.BadStoreConfiguration(
+                store_name="vmware_datastore", reason=msg)
+        self.session = self.reset_session()
+        super(Store, self).configure(re_raise_bsc=re_raise_bsc)
 
     def _get_datacenter(self, datacenter_path):
         search_index_moref = self.session.vim.service_content.searchIndex
@@ -362,7 +443,7 @@ class Store(glance_store.Store):
 
     def _parse_datastore_info_and_weight(self, datastore):
         weight = 0
-        parts = map(lambda x: x.strip(), datastore.rsplit(":", 2))
+        parts = [part.strip() for part in datastore.rsplit(":", 2)]
         if len(parts) < 2:
             msg = _('vmware_datastores format must be '
                     'datacenter_path:datastore_name:weight or '
@@ -413,21 +494,13 @@ class Store(glance_store.Store):
         return ds_map
 
     def configure_add(self):
-        if self.conf.glance_store.vmware_datastores:
-            datastores = self.conf.glance_store.vmware_datastores
-        else:
-            # Backwards compatibility for vmware_datastore_name and
-            # vmware_datacenter_path.
-            datacenter_path = self.conf.glance_store.vmware_datacenter_path
-            datastore_name = self._option_get('vmware_datastore_name')
-            datastores = ['%s:%s:%s' % (datacenter_path, datastore_name, 0)]
-
+        datastores = self._option_get('vmware_datastores')
         self.datastores = self._build_datastore_weighted_map(datastores)
         self.store_image_dir = self.conf.glance_store.vmware_store_image_dir
 
     def select_datastore(self, image_size):
         """Select a datastore with free space larger than image size."""
-        for k, v in sorted(six.iteritems(self.datastores), reverse=True):
+        for k, v in sorted(self.datastores.items(), reverse=True):
             max_ds = None
             max_fs = 0
             for ds in v:
@@ -455,14 +528,15 @@ class Store(glance_store.Store):
     def _build_vim_cookie_header(self, verify_session=False):
         """Build ESX host session cookie header."""
         if verify_session and not self.session.is_current_session_active():
-            self.reset_session(force=True)
+            self.reset_session()
         vim_cookies = self.session.vim.client.options.transport.cookiejar
         if len(list(vim_cookies)) > 0:
             cookie = list(vim_cookies)[0]
             return cookie.name + '=' + cookie.value
 
     @capabilities.check
-    def add(self, image_id, image_file, image_size, context=None):
+    def add(self, image_id, image_file, image_size, context=None,
+            verifier=None):
         """Stores an image file with supplied identifier to the backend
         storage system and returns a tuple containing information
         about the stored image.
@@ -470,24 +544,23 @@ class Store(glance_store.Store):
         :param image_id: The opaque image identifier
         :param image_file: The image data to write, as a file-like object
         :param image_size: The size of the image data to write, in bytes
+        :param verifier: An object used to verify signatures for images
         :retval tuple of URL in backing store, bytes written, checksum
                 and a dictionary with storage system specific information
-        :raises `glance.common.exceptions.Duplicate` if the image already
+        :raises: `glance.common.exceptions.Duplicate` if the image already
                 existed
                 `glance.common.exceptions.UnexpectedStatus` if the upload
                 request returned an unexpected status. The expected responses
                 are 201 Created and 200 OK.
         """
         ds = self.select_datastore(image_size)
+        image_file = _Reader(image_file, verifier)
+        headers = {}
         if image_size > 0:
-            headers = {'Content-Length': image_size}
-            image_file = _Reader(image_file)
+            headers.update({'Content-Length': six.text_type(image_size)})
+            data = image_file
         else:
-            # NOTE (arnaud): use chunk encoding when the image is still being
-            # generated by the server (ex: stream optimized disks generated by
-            # Nova).
-            headers = {'Transfer-Encoding': 'chunked'}
-            image_file = _ChunkReader(image_file)
+            data = utils.chunkiter(image_file, CHUNKSIZE)
         loc = StoreLocation({'scheme': self.scheme,
                              'server_host': self.server_host,
                              'image_dir': self.store_image_dir,
@@ -496,42 +569,47 @@ class Store(glance_store.Store):
                              'image_id': image_id}, self.conf)
         # NOTE(arnaud): use a decorator when the config is not tied to self
         cookie = self._build_vim_cookie_header(True)
-        headers = dict(headers.items() + {'Cookie': cookie}.items())
-        conn_class = self._get_http_conn_class()
-        conn = conn_class(loc.server_host)
-        url = urlparse.quote('%s?%s' % (loc.path, loc.query))
+        headers = dict(headers)
+        headers.update({'Cookie': cookie})
+        session = new_session(self.api_insecure, self.ca_file)
+
+        url = loc.https_url
         try:
-            conn.request('PUT', url, image_file, headers)
+            response = session.put(url, data=data, headers=headers)
         except IOError as e:
+            # TODO(sigmavirus24): Figure out what the new exception type would
+            # be in requests.
             # When a session is not authenticated, the socket is closed by
-            # the server after sending the response. httplib has an open
+            # the server after sending the response. http_client has an open
             # issue with https that raises Broken Pipe
             # error instead of returning the response.
             # See http://bugs.python.org/issue16062. Here, we log the error
             # and continue to look into the response.
-            msg = _LE('Communication error sending http %(method)s request'
+            msg = _LE('Communication error sending http %(method)s request '
                       'to the url %(url)s.\n'
                       'Got IOError %(e)s') % {'method': 'PUT',
                                               'url': url,
                                               'e': e}
             LOG.error(msg)
+            raise exceptions.BackendException(msg)
         except Exception:
             with excutils.save_and_reraise_exception():
                 LOG.exception(_LE('Failed to upload content of image '
                                   '%(image)s'), {'image': image_id})
-        res = conn.getresponse()
-        if res.status == httplib.CONFLICT:
+
+        res = response.raw
+        if res.status == requests.codes.conflict:
             raise exceptions.Duplicate(_("Image file %(image_id)s already "
                                          "exists!") %
                                        {'image_id': image_id})
 
-        if res.status not in (httplib.CREATED, httplib.OK):
+        if res.status not in (requests.codes.created, requests.codes.ok):
             msg = (_LE('Failed to upload content of image %(image)s. '
                        'The request returned an unexpected status: %(status)s.'
                        '\nThe response body:\n%(body)s') %
                    {'image': image_id,
                     'status': res.status,
-                    'body': res.body})
+                    'body': getattr(res, 'body', None)})
             LOG.error(msg)
             raise exceptions.BackendException(msg)
 
@@ -554,7 +632,7 @@ class Store(glance_store.Store):
 
             def another(self):
                 try:
-                    return self.wrapped.next()
+                    return next(self.wrapped)
                 except StopIteration:
                     return ''
 
@@ -567,16 +645,24 @@ class Store(glance_store.Store):
         :param location: `glance_store.location.Location` object, supplied
                         from glance_store.location.get_location_from_uri()
         """
-        return self._query(location, 'HEAD')[2]
+        conn = None
+        try:
+            conn, resp, size = self._query(location, 'HEAD')
+            return size
+        finally:
+            # NOTE(sabari): Close the connection as the request was made with
+            # stream=True.
+            if conn is not None:
+                conn.close()
 
     @capabilities.check
     def delete(self, location, context=None):
         """Takes a `glance_store.location.Location` object that indicates
         where to find the image file to delete
 
-        :location `glance_store.location.Location` object, supplied
+        :param location: `glance_store.location.Location` object, supplied
                   from glance_store.location.get_location_from_uri()
-        :raises NotFound if image does not exist
+        :raises: NotFound if image does not exist
         """
         file_path = '[%s] %s' % (
             location.store_location.datastore_name,
@@ -592,37 +678,66 @@ class Store(glance_store.Store):
             self.session.wait_for_task(delete_task)
         except vexc.FileNotFoundException:
             msg = _('Image file %s not found') % file_path
-            LOG.warn(msg)
+            LOG.warning(msg)
             raise exceptions.NotFound(message=msg)
         except Exception:
             with excutils.save_and_reraise_exception():
                 LOG.exception(_LE('Failed to delete image %(image)s '
                                   'content.') % {'image': location.image_id})
 
-    def _query(self, location, method, depth=0):
-        if depth > MAX_REDIRECTS:
+    def _query(self, location, method):
+        session = new_session(self.api_insecure, self.ca_file)
+        loc = location.store_location
+        redirects_followed = 0
+        # TODO(sabari): The redirect logic was added to handle cases when the
+        # backend redirects http url's to https. But the store never makes a
+        # http request and hence this can be safely removed.
+        while redirects_followed < MAX_REDIRECTS:
+            conn, resp = self._retry_request(session, method, location)
+
+            # NOTE(sigmavirus24): _retry_request handles 4xx and 5xx errors so
+            # if the response is not a redirect, we can return early.
+            if not conn.is_redirect:
+                break
+
+            redirects_followed += 1
+
+            location_header = conn.headers.get('location')
+            if location_header:
+                if resp.status not in (301, 302):
+                    reason = (_("The HTTP URL %(path)s attempted to redirect "
+                                "with an invalid %(status)s status code.")
+                              % {'path': loc.path, 'status': resp.status})
+                    LOG.info(reason)
+                    raise exceptions.BadStoreUri(message=reason)
+                conn.close()
+                location = self._new_location(location, location_header)
+        else:
+            # NOTE(sigmavirus24): We exceeded the maximum number of redirects
             msg = ("The HTTP URL exceeded %(max_redirects)s maximum "
                    "redirects.", {'max_redirects': MAX_REDIRECTS})
             LOG.debug(msg)
             raise exceptions.MaxRedirectsExceeded(redirects=MAX_REDIRECTS)
+
+        content_length = int(resp.getheader('content-length', 0))
+
+        return (conn, resp, content_length)
+
+    def _retry_request(self, session, method, location):
         loc = location.store_location
         # NOTE(arnaud): use a decorator when the config is not tied to self
         for i in range(self.api_retry_count + 1):
             cookie = self._build_vim_cookie_header()
             headers = {'Cookie': cookie}
-            try:
-                conn = self._get_http_conn(method, loc, headers)
-                resp = conn.getresponse()
-            except Exception:
-                with excutils.save_and_reraise_exception():
-                    LOG.exception(_LE('Failed to access image %(image)s '
-                                      'content.') % {'image':
-                                                     location.image_id})
+            conn = session.request(method, loc.https_url, headers=headers,
+                                   stream=True)
+            resp = conn.raw
+
             if resp.status >= 400:
-                if resp.status == httplib.UNAUTHORIZED:
-                    self.reset_session(force=True)
+                if resp.status == requests.codes.unauthorized:
+                    self.reset_session()
                     continue
-                if resp.status == httplib.NOT_FOUND:
+                if resp.status == requests.codes.not_found:
                     reason = _('VMware datastore could not find image at URI.')
                     LOG.info(reason)
                     raise exceptions.NotFound(message=reason)
@@ -631,34 +746,35 @@ class Store(glance_store.Store):
                 LOG.debug(msg)
                 raise exceptions.BadStoreUri(msg)
             break
-        location_header = resp.getheader('location')
-        if location_header:
-            if resp.status not in (301, 302):
-                reason = (_("The HTTP URL %(path)s attempted to redirect "
-                            "with an invalid %(status)s status code.")
-                          % {'path': loc.path, 'status': resp.status})
-                LOG.info(reason)
-                raise exceptions.BadStoreUri(message=reason)
-            location_class = glance_store.location.Location
-            new_loc = location_class(location.store_name,
-                                     location.store_location.__class__,
-                                     uri=location_header,
-                                     image_id=location.image_id,
-                                     store_specs=location.store_specs)
-            return self._query(new_loc, method, depth + 1)
-        content_length = int(resp.getheader('content-length', 0))
+        return conn, resp
 
-        return (conn, resp, content_length)
+    def _new_location(self, old_location, url):
+        store_name = old_location.store_name
+        store_class = old_location.store_location.__class__
+        image_id = old_location.image_id
+        store_specs = old_location.store_specs
+        # Note(sabari): The redirect url will have a scheme 'http(s)', but the
+        # store only accepts url with scheme 'vsphere'. Thus, replacing with
+        # store's scheme.
+        parsed_url = urlparse.urlparse(url)
+        new_url = parsed_url._replace(scheme='vsphere')
+        vsphere_url = urlparse.urlunparse(new_url)
+        return glance_store.location.Location(store_name,
+                                              store_class,
+                                              self.conf,
+                                              uri=vsphere_url,
+                                              image_id=image_id,
+                                              store_specs=store_specs)
 
-    def _get_http_conn(self, method, loc, headers, content=None):
-        conn_class = self._get_http_conn_class()
-        conn = conn_class(loc.server_host)
-        url = urlparse.quote('%s?%s' % (loc.path, loc.query))
-        conn.request(method, url, content, headers)
 
-        return conn
-
-    def _get_http_conn_class(self):
-        if self.api_insecure:
-            return httplib.HTTPConnection
-        return httplib.HTTPSConnection
+def new_session(insecure=False, ca_file=None, total_retries=None):
+    session = requests.Session()
+    if total_retries is not None:
+        http_adapter = adapters.HTTPAdapter(
+            max_retries=retry.Retry(total=total_retries))
+        https_adapter = adapters.HTTPAdapter(
+            max_retries=retry.Retry(total=total_retries))
+        session.mount('http://', http_adapter)
+        session.mount('https://', https_adapter)
+    session.verify = ca_file if ca_file else not insecure
+    return session
